(* absyntofol/absyntofol.sml *)

functor AbsynToFOLFn(structure StrDict : STR_DICT
		     structure Util : UTIL
		     structure Absyn : ABSYN
		     structure FOL : FOL
		       ) : ABSYNTOFOL =
  struct

    structure Absyn = Absyn
    structure FOL = FOL

    fun newvars i =
      let fun loop(i, vars) =
	    if i < 1 then vars else loop(i-1, FOL.newvar() :: vars)
      in
	loop(i, [])
      end

    fun cnvId id = Absyn.identName id
    fun cnvModidOpt(NONE) = NONE
      | cnvModidOpt(SOME modid) = SOME(cnvId modid)
    fun cnvLongid(Absyn.LONGID(modidOpt,id)) =
      let val modidOpt = cnvModidOpt modidOpt
	  val id = cnvId id
      in
	FOL.LONGID(modidOpt, id)
      end
    fun cnvLongidOpt(NONE) = NONE
      | cnvLongidOpt(SOME longid) = SOME(cnvLongid longid)

    fun cnvLit(Absyn.CCONlit c) = FOL.ICON(Char.ord c)
      | cnvLit(Absyn.ICONlit i) = FOL.ICON i
      | cnvLit(Absyn.RCONlit r) = FOL.RCON r
      | cnvLit(Absyn.SCONlit s) = FOL.SCON s

    fun makeVRef(longid as Absyn.LONGID(modidOpt,id), VE) =
      case modidOpt
	of SOME _ => FOL.GVAR(cnvLongid longid)
	 | NONE =>
	    case StrDict.find(VE, cnvId id)
	      of SOME var => FOL.BVAR var
	       | NONE => FOL.GVAR(cnvLongid longid)

    fun cnvExp(Absyn.LITexp lit, _) = FOL.LITexp(cnvLit lit)
      | cnvExp(Absyn.CONexp longcon, _) = FOL.CONexp(cnvLongid longcon)
      | cnvExp(Absyn.VARexp longvar, VE) = FOL.VARexp(makeVRef(longvar, VE))
      | cnvExp(Absyn.STRUCTexp(longconOpt,exps), VE) =
	  FOL.STRUCTexp(cnvLongidOpt longconOpt, cnvExps(exps, VE))
      | cnvExp(Absyn.IDENTexp(_,r), VE) = cnvExp(!r, VE)

    and cnvExps(exps, VE) = map (fn exp => cnvExp(exp, VE)) exps

    fun cnvPat(pat, VE) =
      let val var = FOL.newvar()
	  val (pat', VE') = cnvPat'(pat, var, VE)
      in
	(FOL.PAT(var, pat'), VE')
      end

    and cnvPat'(Absyn.WILDpat, _, VE) = (FOL.WILDpat, VE)
      | cnvPat'(Absyn.LITpat lit, _, VE) = (FOL.LITpat(cnvLit lit), VE)
      | cnvPat'(Absyn.CONpat longcon, _, VE) = (FOL.CONpat(cnvLongid longcon), VE)
      | cnvPat'(Absyn.STRUCTpat(longconOpt,pats), _, VE) =
	  let val (pats', VE') = cnvPats(pats, VE)
	      val longconOpt = cnvLongidOpt longconOpt
	  in
	    (FOL.STRUCTpat(longconOpt, pats'), VE')
	  end
      | cnvPat'(Absyn.BINDpat(id,pat), var, VE) =
	  cnvPat'(pat, var, StrDict.insert(VE, cnvId id, var))
      | cnvPat'(Absyn.IDENTpat(_,r), var, VE) = cnvPat'(!r, var, VE)

    and cnvPats(pats, VE) =
      let fun loop([], pats', VE) = (List.rev pats', VE)
	    | loop(pat::pats, pats', VE) =
		let val (pat', VE) = cnvPat(pat, VE)
		in
		  loop(pats, pat'::pats', VE)
		end
      in
	loop(pats, [], VE)
      end

    fun cnvGoal(Absyn.CALLgoal(longvar,exps,pats), VE) =
	  let val vref = makeVRef(longvar, VE)
	      val exps' = cnvExps(exps, VE)
	      val (pats', VE) = cnvPats(pats, VE)
	      and vars = newvars(List.length pats)
	  in
	    (FOL.AND(FOL.CALL(vref, exps', vars),
		     FOL.MATCH(ListPair.zip(vars, pats'))),
	     VE)
	  end
      | cnvGoal(Absyn.EQUALgoal(id,exp), VE) =
	  let val exp = cnvExp(exp, VE)
	      val id = cnvId id
	  in
	    case StrDict.find(VE, id)	(*XXX: implicit let crap*)
	      of SOME var => (FOL.EQUAL(var, exp), VE)
	       | NONE =>
		  let val var = FOL.newvar()
		  in
		    (FOL.BIND(var, exp), StrDict.insert(VE, id, var))
		  end
	  end
      | cnvGoal(Absyn.LETgoal(pat,exp), VE) =
	  let val exp' = cnvExp(exp, VE)
	      val (pat', VE) = cnvPat(pat, VE)
	      val var = FOL.newvar()
	  in
	    (FOL.AND(FOL.BIND(var, exp'), FOL.MATCH([(var, pat')])), VE)
	  end
      | cnvGoal(Absyn.NOTgoal g, VE) =
	  let val (conj, VE) = cnvGoal(g, VE)
	  in
	    (FOL.NOT conj, VE)
	  end
      | cnvGoal(Absyn.ANDgoal(g1,g2), VE) =
	  let val (conj1, VE) = cnvGoal(g1, VE)
	      val (conj2, VE) = cnvGoal(g2, VE)
	  in
	    (FOL.AND(conj1, conj2), VE)
	  end

    fun cnvClause vars =
      let fun mkdisj(Absyn.CLAUSE1(goalOpt,_,pats,result)) =
		let val (pats', VE) = cnvPats(pats, StrDict.empty)
		    fun return VE =
		      case result
			of Absyn.RETURN exps => FOL.RETURN(cnvExps(exps, VE))
			 | Absyn.FAIL => FOL.FAIL
		    val body = case goalOpt
				 of SOME goal =>
				      let val (conj, VE) = cnvGoal(goal, VE)
				      in
					FOL.ANDTHEN(conj, return VE)
				      end
				  | NONE => return VE
		in
		  FOL.ANDTHEN(FOL.MATCH(ListPair.zip(vars, pats')), body)
		end
	    | mkdisj(Absyn.CLAUSE2(cl1,cl2)) = FOL.ORELSE(mkdisj cl1, mkdisj cl2)
      in
	mkdisj
      end

    fun clauseArity(Absyn.CLAUSE1(_,_,pats,_)) = List.length pats
      | clauseArity(Absyn.CLAUSE2(cl1,_)) = clauseArity cl1

    fun cnvRelBind (Absyn.RELBIND(var,_,cl)) =
      let val vars = newvars(clauseArity cl)
      in
	FOL.REL(cnvId var, vars, cnvClause vars cl)
      end

    fun cnvConBind(Absyn.CONcb con) = FOL.CONcd(cnvId con)
      | cnvConBind(Absyn.CTORcb(con,tys)) = FOL.CTORcd(cnvId con, List.length tys)
    fun cnvDatBind(Absyn.DATBIND(_,tycon,conbinds)) =
      FOL.DATDESC(cnvId tycon, map cnvConBind conbinds)

    fun cnvSpecs specs =
      let fun cnvSpec(spec, specs) =
	    case spec
	      of Absyn.WITHspec(_,ri) => FOL.WITHspec(cnvInterface(!ri))::specs
	       | Absyn.ABSTYPEspec _ => specs
	       | Absyn.TYPEspec _ => specs
	       | Absyn.DATAspec(datbinds,_) =>
		  FOL.DATAspec(map cnvDatBind datbinds)::specs
	       | Absyn.VALspec(var,_) => FOL.VALspec(cnvId var)::specs
	       | Absyn.RELspec(var,_) => FOL.RELspec(cnvId var)::specs
      in
	List.rev(List.foldl cnvSpec [] specs)
      end

    and cnvInterface(Absyn.INTERFACE{modid,specs,...}) =
      FOL.INTERFACE(cnvId modid, cnvSpecs specs)

    fun cnvDecs decs =
      let fun cnvDec(dec, decs) =
	    case dec
	      of Absyn.WITHdec(_,ri) => FOL.WITHdec(cnvInterface(!ri))::decs
	       | Absyn.TYPEdec _ => decs
	       | Absyn.DATAdec(datbinds,_) =>
		  FOL.DATAdec(map cnvDatBind datbinds)::decs
	       | Absyn.VALdec(var,exp) =>
		  FOL.VALdec(cnvId var, cnvExp(exp, StrDict.empty))::decs
	       | Absyn.RELdec relbinds =>
		  FOL.RELdec(map cnvRelBind relbinds)::decs
      in
	List.rev(List.foldl cnvDec [] decs)
      end

    fun translate(Absyn.MODULE(interface,decs)) =
      FOL.MODULE(cnvInterface interface, cnvDecs decs)

  end (* functor AbsynToFOLFn *)
